// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

// This file includes all of the pieces of the BST implementation

// Include the node implementation
#include "BSTNode.h"

// Include the dictionary ADT
#include "dictionary.h"

// Binary Search Tree implementation for the Dictionary ADT
template <typename Key, typename E>
class BST: public Dictionary<Key, E>
{
    private:
        BSTNode<Key, E> *root;  // Root of the BST
        int nodecount;         // Number of nodes in the BST

        // Private "helper" functions
        void clearhelp(BSTNode<Key, E> *);

        BSTNode<Key, E> * inserthelp(BSTNode<Key, E> *, const Key &, const E &);

        BSTNode<Key, E> * inserthelp2(BSTNode<Key, E> *&, const Key &, const E &);

        BSTNode<Key, E> * deletemin(BSTNode<Key, E> *);

        BSTNode<Key, E> * getmin(BSTNode<Key, E> *);

        BSTNode<Key, E> * removehelp(BSTNode<Key, E> *, const Key &);

        E findhelp(BSTNode<Key, E> *, const Key &) const;
        
        int size_help(BSTNode<Key, E> *);

        void printhelp(BSTNode<Key, E> *, string) const;

    public:
        // Constructor
        BST()
        {
            root = nullptr;
            nodecount = 0;
        }

        // Destructor
        ~BST()
        {
            clearhelp(root);
        }

        void clear()   // Reinitialize tree
        {
            clearhelp(root);
            root = nullptr;
            nodecount = 0;
        }

        // Insert a record into the tree.
        // k Key value of the record.
        // e The record to insert.
        void insert(const Key &k, const E &e)
        {
            // Also works
            root = inserthelp(root, k, e);
            //root = inserthelp2(root, k, e);
            nodecount++;
        }

        // Remove a record from the tree.
        // k Key value of record to remove.
        // Return: The record removed, or nullptr if there is none.
        E remove(const Key &k)
        {
            E temp = findhelp(root, k);   // First find it

            if(temp != nullptr)
            {
                root = removehelp(root, k);
                nodecount--;
            }

            return temp;
        }

        // Remove and return the root node from the dictionary.
        // Return: The record removed, null if tree is empty.
        // Useful if you just want to read out every item semi-randomly.
        E removeAny()
        {
            if(root != nullptr)
            {
                E temp = root->element();
                root = removehelp(root, root->key());
                nodecount--;
                return temp;
            }
            else
                return nullptr;
        }

        // Return Record with key value k, nullptr if none exist.
        // k: The key value to find. */
        // Return some record matching "k".
        // Return true if such exists, false otherwise. If
        // multiple records match "k", return an arbitrary one.
        E find(const Key &k) const
        {
            return findhelp(root, k);
        }

        // Return the number of records in the dictionary.
        int size()
        {
            return nodecount;
        }
        
        int size2() 
        {
            return size_help(root);
        }

        void print() const
        {
            cout << "============== Printing tree ==============" << endl;
            // Print the contents of the BST
            if(root == nullptr)
                cout << "The BST is empty" << endl;
            else
                printhelp(root, "");
        }
};

template <typename Key, typename E>
int BST<Key, E>::size_help(BSTNode<Key, E> *root) 
{ 
    // if Nothing to count 
    if(root == nullptr) 
        return 0;  
    return 1 + size_help(root->left()) + size_help(root->right()); 
}


// Clean up BST by releasing space back free store
// What happens if this is pre-order??
template <typename Key, typename E>
void BST<Key, E>::clearhelp(BSTNode<Key, E> *root)
{
    if(root == nullptr)
        return;

    clearhelp(root->left());
    clearhelp(root->right());
    delete root;
}


// Insert a node into the BST, returning the updated tree
// With ref, could be made more efficient by modifying root and omitting setLeft/Right
template <typename Key, typename E>
BSTNode<Key, E> * BST<Key, E>::inserthelp(BSTNode<Key, E> *root, const Key &k, const E &it)
{
    if(root == nullptr)  // Empty tree: create node
        return new BSTNode<Key, E>(k, it, nullptr, nullptr);
    else if(k < root->key())
        root->setLeft(inserthelp(root->left(), k, it));
    else
        root->setRight(inserthelp(root->right(), k, it));

    // Return pointer to root of tree with node inserted
    return root;
}


template <typename Key, typename E>
BSTNode<Key, E> * BST<Key, E>::inserthelp2(BSTNode<Key, E> *&root, const Key &k, const E &it)
{
    if(root == nullptr)  // Empty tree: create node
        root = new BSTNode<Key, E>(k, it, nullptr, nullptr);
    else if(k < root->key())
    {
        inserthelp2(root->left2(), k, it);
        //inserthelp2(root->lc, k, it); // if public
    }
    else
        inserthelp2(root->right2(), k, it);

    //inserthelp2(root->rc, k, it);
    return root;       // Return tree with node inserted
}


// Delete the minimum value from the BST, returning the revised BST
template <typename Key, typename E>
BSTNode<Key, E> * BST<Key, E>::getmin(BSTNode<Key, E> *rt)
{
    if(rt->left() == nullptr)
        return rt;
    else
        return getmin(rt->left());
}


// With ref could be made more efficient by modifying rt and omitting setLeft/Right, like above
template <typename Key, typename E>
BSTNode<Key, E> * BST<Key, E>::deletemin(BSTNode<Key, E> *rt)
{
    if(rt->left() == nullptr) // Found min
        return rt->right();
    else
    {
        // Continue left
        rt->setLeft(deletemin(rt->left()));
        return rt;
    }
}


// Remove a node with key value k
// Return: The tree with the node removed
template <typename Key, typename E>
BSTNode<Key, E> * BST<Key, E>::removehelp(BSTNode<Key, E> *rt, const Key &k)
{
    if(rt == nullptr) // k is not in tree -- never needed due to check above?
        return nullptr;
    else if(k < rt->key())
        rt->setLeft(removehelp(rt->left(), k));
    else if(k > rt->key())
        rt->setRight(removehelp(rt->right(), k));
    else
    {
        // Found: remove it
        BSTNode<Key, E> *temp = rt;

        if(rt->left() == nullptr) // Only a right child (at max) Also handles no children
        {
            rt = rt->right();         //  so point to right
        }
        else if(rt->right() == nullptr) // Only a left child (at max) Also handles no children
        {
            rt = rt->left();          //  so point to left
        }
        else // Both children are present
        {
            BSTNode<Key, E> *temp = getmin(rt->right());
            rt->setElement(temp->element());
            rt->setKey(temp->key());
            rt->setRight(deletemin(rt->right()));
        }

        delete temp;
    }

    return rt;
}


// Find a node with the given key value
template <typename Key, typename E>
E BST<Key, E>::findhelp(BSTNode<Key, E> *root, const Key &k) const
{
    if(root == nullptr)
        return nullptr;          // Empty tree

    if(k < root->key())
        return findhelp(root->left(), k);   // Check left
    else if(k > root->key())
        return findhelp(root->right(), k);  // Check right
    else
        return root->element();  // Found it
}


// Print out a BST
template <typename Key, typename E>
void BST<Key, E>::printhelp(BSTNode<Key, E> *root, string indent) const
{
    if(root == nullptr)// Empty tree
    {
        cout << indent << "   Empty" << endl;
        return;
    }

    printhelp(root->right(), indent += "  "); // Do right subtree
    cout << indent << " " << root->key() << endl;        // Print node value
    printhelp(root->left(), indent += ""); // Do left subtree
}




